/*
 *     TERMCAP
 */

#include "head.h"
#include <strings.h>
#ifndef __MSDOS__
#ifndef __WIN32__
#include <unistd.h>
#include <sys/ioctl.h>
#endif
#endif

struct TTelem {
  char *Name;
  void *ptr;
};

int sc_width=80, sc_height=25;
void scrsize();

char term[256], oldterm[256];
char *se, *so, *us;
char *T_normal,*T_highlight,*T_sel,*T_special,*T_head,
     *T_control,*T_error,*T_dialog,*T_button,
     *T_virgin,*T_line,*T_cline,*T_sline,*T_number,*T_cursor;

static TTelem table[]={
  {"tc",&term}, //    tc!
  {"co",&sc_width},
  {"li",&sc_height},
  {"cm",&_curpos},
  {"ce",&_clreol},
  {"cl",&_clrscr},
  {"al",&_iline},
  {"dl",&_dline},
  {"sr",&_scrup},
  {"sf",&_scrdn},
  {"cs",&_scra},
  {"se",&se},
  {"so",&so},
  {"us",&us},
  {"is",&_is},
  {"rs",&_rs},
  {"Xt",&_xtitle},
  {"ac",&_graph},
  {"gt",&_graph0},
  {"as",&_graph_on},
  {"ae",&_graph_off},
  {"bl",&_bell},
  {"pc",&_fill},
  {"pb",&minspeed},
  {"Ct",&_rusmap},
  {"Cs",&_rus},
  {"Ce",&_lat},
  {"CY",&_isrus},

  {"me",&_std},
  {"No",&T_normal},
  {"Mr",&T_highlight},
  {"So",&T_sel},
  {"Sp",&T_special},
  {"Hd",&T_head},
  {"Cl",&T_control},
  {"Er",&T_error},
  {"Di",&T_dialog},
  {"Bu",&T_button},
  {"Vi",&T_virgin},
  {"Li",&T_line},
  {"Ci",&T_cline},
  {"Si",&T_sline},
  {"Nu",&T_number},
  {"Cu",&T_cursor},

  {"k1",&syms0[1].str},
  {"k2",&syms0[2].str},
  {"k3",&syms0[3].str},
  {"k4",&syms0[4].str},
  {"k5",&syms0[5].str},
  {"k6",&syms0[6].str},
  {"k7",&syms0[7].str},
  {"k8",&syms0[8].str},
  {"k9",&syms0[9].str},
  {"k;",&syms0[10].str},

  {"s1",&syms0[11].str},
  {"s2",&syms0[12].str},
  {"s3",&syms0[13].str},
  {"s4",&syms0[14].str},
  {"s5",&syms0[15].str},
  {"s6",&syms0[16].str},
  {"s7",&syms0[17].str},
  {"s8",&syms0[18].str},
  {"s9",&syms0[19].str},
  {"s0",&syms0[20].str},

  {"a1",&syms0[21].str},
  {"a2",&syms0[22].str},
  {"a3",&syms0[23].str},
  {"a4",&syms0[24].str},
  {"a5",&syms0[25].str},
  {"a6",&syms0[26].str},
  {"a7",&syms0[27].str},
  {"a8",&syms0[28].str},
  {"a9",&syms0[29].str},
  {"a0",&syms0[30].str},

  {"ku",&syms0[31].str},
  {"kd",&syms0[32].str},
  {"kr",&syms0[33].str},
  {"kl",&syms0[34].str},
  {"kI",&syms0[35].str},
  {"kh",&syms0[36].str},
  {"@7",&syms0[37].str},
  {"kP",&syms0[38].str},
  {"kN",&syms0[39].str},
  {"kD",&syms0[40].str},
  {"kb",&syms0[41].str},

  {"",NULL}
}; //table

FILE *stream;
char *sbuffer;

int StrCmp (const char *s1, const char*s2) {
  if (s1==NULL) return -1;
  if (s2==NULL) return 1;
  return strcmp(s1,s2);
}

int StrNCmp (const char *s1, const char *s2, size_t maxlen) {
  if (s1==NULL) return -1;
  if (s2==NULL) return 1;
  return strncmp(s1,s2,maxlen);
}

int StrLen (const char *s) {
  if (s==NULL) return 0;
  return strlen(s);
}

int end_stream() {
  if (stream) return feof(stream);
  if (sbuffer) return *sbuffer==0;
  return 1;
} //end_stream

void open_stream (char *name) {
  stream=NULL; sbuffer=NULL;
  if (name==NULL) return;
#ifndef __MSDOS__
  if (*name=='/')
#endif
  stream=fopen(name,"r");
  if (stream==NULL)
    sbuffer=name;
} //open_stream;

void close_stream() {
  if (stream) fclose(stream);
  if (sbuffer) sbuffer=NULL;
} //close_stream

unsigned gc() {
  if (stream) {
    if (feof(stream)) return 0;
    return getc(stream);
  }
  else {
    if ((sbuffer==NULL) || (*sbuffer==0)) return 0;
    return *(sbuffer++);
  }
} //gc

void ugc (unsigned c) {
  if (stream)
    ungetc(c,stream);
  else if (sbuffer)
    *(--sbuffer)=c;
} //ugc

const unsigned EOD= ~0;

unsigned inchar() {
  unsigned c;

  c=gc();
  if (c==0) return 0;
retry:
  switch (c) {
    case '^':
      c=gc();
      return toupper(c)-0100;
    case ':':
      return EOD;
    case '\n':
      return 0;
    case '\\':
      c=tolower(gc());
      switch (c) {
        case '\\':  return '\\';
        case '^':   return '^';
        case 'e':   return 033;
        case 'n':   return 012;
        case 'r':   return 015;
        case 't':   return 011;
        case 'b':   return 010;
        case 'f':   return 014;
        case '\n':
          do c=gc();
          while ((c==' ')||(c=='\t'));
          goto retry;
        default:
          if ((c<'0')||(c>'7')) return c;
          unsigned num=(c-'0');
          c=gc();
          if ((c<'0')||(c>'7')) {
            ugc(c); return num;
          }
          num=num*8+(c-'0');
          c=gc();
          if ((c<'0')||(c>'7')) {
            ugc(c); return num;
          }
          num=num*8+(c-'0');
          return num;
      } //switch 2
    default:
      return c;
  } //switch
} //inchar

unsigned skip() {
  unsigned c;

  do c=inchar();
  while ((c!=0)&&(c!=EOD));
  return c;
} //skip

void getline (char *line, int size) {
  int i;
  unsigned c;

  for(i=0; i<size-1;) {
    c=inchar();
    if ((c==0)||(c==EOD)) break;
    line[i++]=c;
  }
  line[i]=0;
} //getline

void getnum (int &num) {
  unsigned c;

  for (num=0; ;) {
    c=inchar();
    if ((c==0)||(c==EOD)) break;
    num=num*10+(c-'0');
  }
} //getnum

static TTelem *find (char *name, TTelem *tbl) {
  TTelem *p;

  if (name==NULL) return NULL;
  for (p=tbl; StrLen(p->Name); p++)
    if (StrCmp(p->Name,name)==0)
      return p;
  return NULL;
} //find

void analys() {
  char name[3];
  char line[80];
  TTelem *p;
  unsigned c;
  int num;

  term[0]=0;
  table[0].ptr=term; // 
  for (;;) {
    c=inchar();
    if (c==EOD) continue;
    if (c==0) return;
    name[0]=c;
    name[1]=inchar();
    name[2]=0;
    c=inchar();
    switch (c) {
      case EOD:     //  
        p=find(name,table);
        if (p==NULL) break;
        if (p->ptr==NULL) break;
        *(int*)(p->ptr)=1; // 
        p->ptr=NULL;
        break;
      case 0:       //  
        return;
      case '=':     //  
        getline(line,sizeof(line));
        p=find(name,table);
        if (p==NULL) break;
        if (p->ptr==NULL) break;
        if (p==&table[0]) { //  tc (term)
          strcpy((char*)(p->ptr),line);
        }
        else {
          *(char**)(p->ptr)=new char[StrLen(line)+1];
          strcpy(*(char**)(p->ptr),line);
          p->ptr=NULL;
        }
        break;
      case '#':     //  
        getnum(num);
        p=find(name,table);
        if (p==NULL) break;
        if (p->ptr==NULL) break;
        *(int*)(p->ptr)=num;
        p->ptr=NULL;
        break;
      case '@':
        p=find(name,table);
        if (p==NULL) break;
        if (p->ptr==NULL) break;
        *(char**)(p->ptr)=NULL;
        p->ptr=NULL;
        skip();
        break;
    } //switch
  } //for
} //analys

int dotermcap (char *termcap) {
  char term0[120];
  char *p;
  char c;

  ///**/fprintf(stderr,"TERM=%s TERMCAP=%s\n",term,termcap);
  if (term[0]==0) return 1;

  open_stream(termcap);
  while (!end_stream()) {
    c=gc();
    if (c==0) break;
    switch (c) {
      case '#':
        while ((c=gc())!='\n');
        break;
      case ' ':
      case '\t':
      case '\n':
        break;
      default:
        for (p=term0; ;p++) {
          *p=c;
          c=gc();
          if ((c=='|')||(c==':')||(c==0)) break;
        }
        *++p=0;
        if (StrCmp(term0,term)!=0) {
          if (c==':')
            while (inchar());
          break;
        }
        if (c=='|')
          while ((c=gc())!=':');
        strcpy(oldterm,term); //  

        analys(); //   termcap

        if (StrCmp(term,"*")==0) {
          strcpy(term,oldterm); //   
          break;
        }
        else if (term[0]) {
          continue;
        }
        else goto fin;
        break;
    } //switch
  } //while
fin:
  close_stream();
  return (term[0]==0);
} //dotermcap

#ifdef __MSDOS__
 char *TERMCAP0="TERMCAP.";
#else
 char *TERMCAP0=_TERMCAP_;
#endif

char RusMap[256];
extern unsigned color;

void check_color() {
  _sel=T_sel;
  _normal=T_normal;
  _highlight=T_highlight;
  _head=T_head;
  _special=T_special;
  _error=T_error;
  _control=T_control;
  _dialog=T_dialog;
  _button=T_button;
  _virgin=T_virgin;
  _line=T_line;
  _cline=T_cline;
  _sline=T_sline;
  _number=T_number;
  _cursor=T_cursor;

  if (!color||_sel==NULL)   _sel=so;
  if (!color||_normal==NULL)  _normal=_std;
  if (!color||_head==NULL)  _head=so;
  if (!color||_special==NULL) _special=us;
  if (!color||_error==NULL) _error=so;
  if (!color||_highlight==NULL) _highlight=_std;
  if (!color||_control==NULL) _control=us;
  if (!color||_dialog==NULL)  _dialog=_sel;
  if (!color||_button==NULL)  _button=_special;
  if (!color||_virgin==NULL)  _virgin=_head;
  if (!color||_line==NULL)  _line=_normal;
  if (!color||_cline==NULL) _cline=_control;
  if (!color||_sline==NULL) _sline=_special;
  if (!color||_number==NULL)  _number=_std;
  if (!color||_cursor==NULL)  _cursor=_sel;
} //check_color

int getac (int c) {
  char *p;
  p = strchr/*index*/(_graph, c);
  if (p) return *++p;
  return 0;
} //getac

int dumbterm() {
  return (_curpos==NULL)||(_clreol==NULL);
} //dumbterm

void gettermcap() {
  char filename[2048];
  char filename2[2048];
  char *pp;
  int i;
  char *ttname;

  ttname=getenv("TERM");
#ifndef __WIN32__
  if (ttname==0 || strlen(ttname)==0) {
    fprintf(stderr,"TERM variable is empty\n");
    _exit(0);
  }
  strcpy(term,ttname);
#endif
  gethome(filename);
#ifdef __WIN32__
  if (strlen(filename)>2 && filename[1]==':') {
    char temp[1024];
    filename[1]='/';
    strcpy(temp,"//");
    strcat(temp,filename);
    strcpy(filename,temp);
  }
#endif

  strcat(filename,"termcap");

  pp=getenv("TERMCAP");
  if (pp && strlen(pp)>0)
    strcpy(filename2,pp);
  else
    strcpy(filename2,_TERMCAP0_);

  for (i=0; i<10; i++) {
    if (dotermcap(filename)) break;
    if (dotermcap(TERMCAP0)) break;
    if (dotermcap(filename2)) break;
  }

  ///**/fprintf(stderr,"DUMB=%d\n",dumbterm());
  ///_exit(0);
  scrsize();

#ifndef __WIN32__
  if ((SWIDTH<10)||(SHEIGHT<5)||dumbterm()) {
    fprintf(stderr,"Invalid TERMCAP or too dumb terminal\n");
    fprintf(stderr,
      "SWIDTH=%d, SHEIGHT=%d, _curpos=%d, _clreol=%d\n",
      SWIDTH, SHEIGHT, _curpos, _clreol);
    _exit(0);
  }
#else
  CONSOLE_SCREEN_BUFFER_INFO info;
  GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE),&info);
  SWIDTH=info.dwSize.X-1;
  SHEIGHT=info.dwSize.Y-1;
  //_scrdn=NULL;
  _iline="";
  _dline="";
#endif

  check_color();

  if (_graph) {
    static char seq[] = "lmkjxxqq";
    int c, ch;
    for (c=0; c<8; c++) {
      ch = getac(seq[c]);
      if (ch) _graph0[c] = ch;
    }
  }
  _graph=_graph0;
  if (_rusmap==NULL) return;

  FILE *F=fopen(_rusmap,"r");
  if (F==NULL) return;
  char *p=RusMap;
  while (!feof(F)) *p++=getc(F);
  fclose(F);

  ///**/fprintf(stderr,"getterm DONE\n");
} //gettermcap

#ifdef __MSDOS__
void scrsize() {
  SWIDTH=sc_width-1; SHEIGHT=sc_height-1;
}
#else
void scrsize() {
  register char *s;
#ifdef TIOCGWINSZ
  struct winsize w;
#else
#ifdef WIOCGETD
  struct uwdata w;
#endif
#endif

#ifdef TIOCGWINSZ
  if (ioctl(2, TIOCGWINSZ, &w) == 0 && w.ws_row > 0)
    sc_height = w.ws_row;
  else
#else
#ifdef WIOCGETD
  if (ioctl(2, WIOCGETD, &w) == 0 && w.uw_height > 0)
    sc_height = w.uw_height/w.uw_vs;
  else
#endif
#endif
  if ((s = getenv("LINES")) != NULL)
    sc_height = atoi(s);
  //else  sc_height = tgetnum("li");

  if (sc_height <= 0)
    sc_height = 24;

#ifdef TIOCGWINSZ
  if (ioctl(2, TIOCGWINSZ, &w) == 0 && w.ws_col > 0)
    sc_width = w.ws_col;
  else
#ifdef WIOCGETD
  if (ioctl(2, WIOCGETD, &w) == 0 && w.uw_width > 0)
    sc_width = w.uw_width/w.uw_hs;
  else
#endif
#endif
  if ((s = getenv("COLUMNS")) != NULL)
    sc_width = atoi(s);
  //else  sc_width = tgetnum("co");

  if (sc_width <= 0)
    sc_width = 80;
  SWIDTH=sc_width-1; SHEIGHT=sc_height-1;
}
#endif
